"""Defines a framework for managing services."""

import cx_Threads
import os
import sys
import time

class ServiceFramework(object):
    """Framework for managing services."""

    def __init__(self):
        self.logger = None
        self.event = cx_Threads.Event()
        self.stopEvent = cx_Threads.Event()
        self.stopRequested = False

    def Initialize(self):
        """Initialize the service in preparation for running. Override this
           method in subclasses as needed."""
        pass

    def Log(self, message, *args):
        """Log a message via the logging thread."""
        self.logger.Log(message % args)

    def Run(self):
        """Run the service. Override this method in subclasses as needed."""
        self.event.Wait()

    def StartLogging(self, logFile):
        """Start logging to the given file; stdout and stderr are redirected
           to this file so that nothing is lost even if an unhandled
           exception is raised. This file is opened unbuffered so that log
           contents are immediately visible. A separate thread is set up to
           handle this so that writes to the file are done in the same way
           in a multithreaded environment."""
        if isinstance(logFile, str):
            logFile = file(logFile, "a", 0)
            try:
                os.dup2(logFile.fileno(), sys.stdout.fileno())
                os.dup2(logFile.fileno(), sys.stderr.fileno())
            except:
                pass
        self.logger = Logger(logFile)
        self.logger.event = self.event
        self.logger.Start()

    def Stop(self):
        """Stop the service. Override this method in subclasses as needed."""
        self.stopRequested = True
        self.event.Set()
        self.stopEvent.Wait()


class Logger(cx_Threads.Queue):
    """Log messages to a file."""

    def __init__(self, logFile):
        self.logFile = logFile
        cx_Threads.Queue.__init__(self)
        self.name = "Logger"

    def Log(self, message):
        """Add a message to the queue of messages to be logged."""
        self.AddItem(message)

    def ProcessItem(self, item):
        """Log the message to the file."""
        timeString = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
        print >> self.logFile, timeString, item
        self.logFile.flush()


class Thread(cx_Threads.Thread):
    """Base class for threads which logs status to a logging thread."""

    def __init__(self, name, event, logger, *args):
        cx_Threads.Thread.__init__(self, self.Run, *args)
        self.name = name
        self.event = event
        self.logger = logger

    def Log(self, message, *args):
        """Log message via the logger."""
        if args:
            message = message % args
        self.logger.Log("%s: %s" % (self.name, message))

    def OnThreadEnd(self):
        """Called when the thread ends."""
        if self.raisedException:
            self.Log("raised exception\n" + self.traceback)
        else:
            self.Log("ending")

    def OnThreadStart(self):
        """Called when the thread is started."""
        self.Log("starting")


class Timer(Thread):
    """Thread which sleeps until the allotted time has expired."""

    def Run(self, timeInSeconds):
        time.sleep(timeInSeconds)

