"""Classes for handling the saving and restoring of settings."""

import os
import sys

if sys.platform == "win32":

    import win32api
    import win32con
    import pywintypes

    class Session:
        """Handles saving and restoring settings from the registry."""

        def __init__(self, registryLoc, baseName):
            self.key = win32api.RegCreateKey(win32con.HKEY_CURRENT_USER,
                    os.path.join(registryLoc, baseName))

        def GetValue(self, key, defaultValue = None):
            """Return a value from the registry, returning the default value
               if a value cannot be found and destroying invalid values if
               any are found."""
            try:
                value, type = win32api.RegQueryValueEx(self.key, key)
            except:
                return defaultValue
            try:
                return eval(str(value))
            except:
                win32api.RegDeleteValue(self.key, key)
                return defaultValue

        def SetValue(self, key, value):
            """Set the value in the registry."""
            win32api.RegSetValueEx(self.key, key, 0, win32con.REG_SZ,
                    repr(value))


    class Handler:
        """Handles saving and restoring settings in the registry."""

        def __init__(self, vendorName, applicationName):
            self.registryLoc = os.path.join("SOFTWARE", vendorName,
                    applicationName)
            baseKey = win32api.RegCreateKey(win32con.HKEY_CURRENT_USER,
                    "SOFTWARE")
            vendorKey = win32api.RegCreateKey(baseKey, vendorName)
            win32api.RegCreateKey(vendorKey, applicationName)

        def GetSession(self, baseName):
            """Return a session which can be used for saving and restoring
               values from the registry. This means that the registry is not
               constantly being opened and closed during a series of getting
               and setting values."""
            return Session(self.registryLoc, baseName)


else:

    import gdbm

    class Session:
        """Handles saving and restoring settings from a GDBM database."""

        def __init__(self, fileName, baseName):
            self.database = gdbm.open(fileName, "cf")
            self.baseName = baseName

        def GetValue(self, key, defaultValue = None):
            """Return a value from the database, returning the default value
               if a value cannot be found and destroying invalid values if
               any are found."""
            key = "%s.%s" % (self.baseName, key)
            try:
                value = self.database[key]
            except KeyError:
                return defaultValue
            try:
                return eval(value)
            except:
                del self.database[key]
                return defaultValue

        def SetValue(self, key, value):
            """Set the value in the database."""
            key = "%s.%s" % (self.baseName, key)
            self.database[key] = repr(value)

    class Handler:
        """Handles saving and restoring settings in a GDBM database."""

        def __init__(self, vendorName, applicationName):
            baseDirName = ".%s" % vendorName.lower()
            baseFileName = "%s.db" % applicationName
            self.fileName = os.path.join(os.environ["HOME"], baseDirName,
                    baseFileName)
            dirName = os.path.dirname(self.fileName)
            if not os.path.isdir(dirName):
                os.makedirs(dirName)

        def GetSession(self, baseName):
            """Return a session which can be used for saving and restoring
               values from the database. This means that the file is not
               constantly being opened and closed during a series of getting
               and setting values."""
            return Session(self.fileName, baseName)

