"""Defines methods for managing threads, events and queues."""

import sys
import thread
import time
import traceback

class Thread:
    """Base class for threads which is a little more lightweight than the
       threading module exports and allows for finer control."""

    def __init__(self, function, *args, **keywordArgs):
        self.function = function
        self.args = args
        self.keywordArgs = keywordArgs
        self.started = False
        self.stopped = False
        self.raisedException = False
        self.traceback = None
        self.event = None
        self.name = None

    def OnThreadEnd(self):
        """Called when the thread is ended. Override in child classes."""
        pass

    def OnThreadStart(self):
        """Called when the thread is started. Override in child classes."""
        pass

    def Start(self):
        """Start the thread."""
        self.started = True
        self.stopped = False
        self.raisedException = False
        self.traceback = None
        thread.start_new_thread(self._Run, ())

    def _Run(self):
        """Execute the function associated with the thread."""
        try:
            self.OnThreadStart()
            try:
                self.function(*self.args, **self.keywordArgs)
            except:
                self.raisedException = True
                lines = traceback.format_exception(*sys.exc_info())
                self.traceback = "".join(lines)
        finally:
            self.stopped = True
            self.OnThreadEnd()
            if self.event:
                self.event.Set()


class Event:
    """Event class which permits synchronization between threads."""

    def __init__(self):
        self.lock = thread.allocate_lock()
        self.isSet = False
        self.waiters = []

    def Clear(self):
        """Clear the flag."""
        self.lock.acquire()
        self.isSet = False
        self.lock.release()

    def Set(self):
        """Set the flag and notify all waiters of the event."""
        self.lock.acquire()
        self.isSet = True
        if self.waiters:
            for waiter in self.waiters:
                waiter.release()
            self.waiters = []
            self.isSet = False
        self.lock.release()

    def Wait(self):
        """Wait for the flag to be set and immediately reset it."""
        self.lock.acquire()
        if self.isSet:
            self.isSet = False
            self.lock.release()
        else:
            waiterLock = thread.allocate_lock()
            waiterLock.acquire()
            self.waiters.append(waiterLock)
            self.lock.release()
            waiterLock.acquire()


class Queue(Thread):
    """Implements a queue which is fed in one thread and consumed in
       another."""

    def __init__(self):
        self.lock = thread.allocate_lock()
        self.stopRequested = False
        self.queueEvent = Event()
        self.items = []
        Thread.__init__(self, self.Run)

    def AddItem(self, item):
        """Add an item to the queue."""
        self.lock.acquire()
        self.items.append(item)
        self.lock.release()
        self.queueEvent.Set()

    def CanProcess(self, requestedItems):
        """Return the number of items that can be processed by the queue at
           this time. This routine is intended to be overridden in
           subclasses."""
        return requestedItems

    def CanStop(self):
        """Return a boolean indicating if the queue can be stopped. This
           routine is intended to be overridden in subclasses."""
        return True

    def ProcessItem(self, item):
        """Process the item in the queue. This routine is intended to be
           overridden in subclasses."""
        pass

    def Run(self):
        """Process items in the queue until a stop is requested."""
        self.lock.acquire()
        self.stopRequested = False
        self.lock.release()
        while True:
            self.lock.acquire()
            numItemsToProcess = self.CanProcess(len(self.items))
            itemsToProcess = self.items[:numItemsToProcess]
            self.items = self.items[numItemsToProcess:]
            self.lock.release()
            for item in itemsToProcess:
                self.ProcessItem(item)
            if self.stopRequested and self.CanStop():
                break
            self.queueEvent.Wait()

    def Stop(self):
        """Request that the queue be stopped."""
        self.lock.acquire()
        self.stopRequested = True
        self.lock.release()
        self.queueEvent.Set()


class ResourcePool:
    """Implements a pool of resources."""

    def __init__(self, maxResources, newResourceFunc):
        self.lock = thread.allocate_lock()
        self.poolEvent = Event()
        self.freeResources = []
        self.busyResources = []
        self.maxResources = maxResources
        self.newResourceFunc = newResourceFunc

    def Destroy(self):
        """Destroy the resource pool, this blocks until all resources are
           returned to the pool for destruction."""
        self.lock.acquire()
        self.freeResources = []
        self.maxResources = 0
        self.lock.release()
        while self.busyResources:
            self.poolEvent.Wait()

    def Get(self):
        """Gets a resource form the pool, creating new resources as necessary.
           The calling thread will block until a resource is available, if
           necessary."""
        resource = None
        self.lock.acquire()
        while resource is None:
            try:
                if self.freeResources:
                    resource = self.freeResources.pop()
                elif len(self.busyResources) < self.maxResources:
                    resource = self.newResourceFunc()
                elif not self.maxResources:
                    raise "No resources not available."
                else:
                    self.lock.release()
                    self.poolEvent.Wait()
                    self.lock.acquire()
            except:
                if self.lock.locked():
                    self.lock.release()
                raise
        self.busyResources.append(resource)
        self.lock.release()
        return resource

    def Put(self, resource, addToFreeList = True):
        """Put a resource back into the pool."""
        self.lock.acquire()
        try:
            index = self.busyResources.index(resource)
            del self.busyResources[index]
            if self.maxResources and addToFreeList:
                self.freeResources.append(resource)
        finally:
            self.lock.release()
        self.poolEvent.Set()


class Timer(Thread):
    """Operates a timer."""

    def __init__(self, timeInSeconds):
        Thread.__init__(self, time.sleep, timeInSeconds)

