#==============================================================================
# cx_XML.py
#   Class which makes parsing or writing of XML files a little simpler.
#==============================================================================

from xml.sax import saxutils

# make freeze work properly
from xml.sax.drivers2 import drv_pyexpat
from xml.parsers import expat

#------------------------------------------------------------------------------
# Parser
#   Parser for the XML file.
#------------------------------------------------------------------------------
class Parser(saxutils.DefaultHandler):

  #----------------------------------------------------------------------------
  # Handlers()
  #   Return a list of handlers for the given handler class.
  #----------------------------------------------------------------------------
  def Handlers(self, a_Class):
    v_Handlers = {}
    for v_Name in self.__class__.__dict__:
      if v_Name.startswith(a_Class):
        v_Handlers[v_Name[len(a_Class):]] = getattr(self, v_Name)
    return v_Handlers

  #----------------------------------------------------------------------------
  # startDocument()
  #   Called when the document is first being processed.
  #----------------------------------------------------------------------------
  def startDocument(self):
    self.i_StartElementHandlers = self.Handlers("start_")
    self.i_EndElementHandlers = self.Handlers("end_")
    self.i_CharactersHandlers = self.Handlers("characters_")

  #----------------------------------------------------------------------------
  # startElement()
  #   Called when an element is started; note that the attributes that are
  # passed are a thin wrapper for a dictionary which contains unicode strings
  # which cannot be used with keyword parameters.
  #----------------------------------------------------------------------------
  def startElement(self, a_Name, a_Attrs):
    v_Handler = self.i_StartElementHandlers.get(a_Name)
    if not v_Handler:
      raise "No start handler for tag '%s'" % str(a_Name)
    v_Attrs = {}
    for v_Key, v_Value in a_Attrs.items():
      v_Attrs[str(v_Key)] = str(v_Value)
    v_Handler(**v_Attrs)
    self.i_ElementName = a_Name

  #----------------------------------------------------------------------------
  # endElement()
  #   Called when an element is ended.
  #----------------------------------------------------------------------------
  def endElement(self, a_Name):
    v_Handler = self.i_EndElementHandlers.get(a_Name)
    if v_Handler:
      v_Handler()

  #----------------------------------------------------------------------------
  # characters()
  #   Called when element values are encountered.
  #----------------------------------------------------------------------------
  def characters(self, a_Data):
    v_Handler = self.i_CharactersHandlers.get(self.i_ElementName)
    if v_Handler:
      v_Handler(str(a_Data))

#------------------------------------------------------------------------------
# Writer
#   Writer for an XML file.
#------------------------------------------------------------------------------
class Writer:

  #----------------------------------------------------------------------------
  # __init__()
  #   Constructor.
  #----------------------------------------------------------------------------
  def __init__(self, a_File):
    self.i_File = a_File
    self.i_Tags = []
    print >> self.i_File, '<?xml version="1.0"?>'

  #----------------------------------------------------------------------------
  # _WriteIndentString()
  #   Write the indent string to the file.
  #----------------------------------------------------------------------------
  def _WriteIndentString(self):
    v_NumSpaces = len(self.i_Tags) * 2
    self.i_File.write(" " * v_NumSpaces)

  #----------------------------------------------------------------------------
  # _WriteTag()
  #   Write the tag with the attributes.
  #----------------------------------------------------------------------------
  def _WriteTag(self, a_Name, a_MultiLineAttributes, a_Attributes):
    self._WriteIndentString()
    self.i_File.write("<%s" % a_Name)
    for v_Key, v_Value in a_Attributes.iteritems():
      if v_Value is None:
        v_Value = ""
      v_Value = saxutils.quoteattr(str(v_Value))
      if a_MultiLineAttributes:
        self.i_File.write('\n        %s=%s' % (v_Key, v_Value))
      else:
        self.i_File.write(' %s=%s' % (v_Key, v_Value))

  #----------------------------------------------------------------------------
  # WriteTagNoValue()
  #   Write a tag which has no value (or any children).
  #----------------------------------------------------------------------------
  def WriteTagNoValue(self, a_Name, **a_Attributes):
    self._WriteTag(a_Name, 0, a_Attributes)
    print >> self.i_File, "/>"

  #----------------------------------------------------------------------------
  # WriteTagWithValue()
  #   Write a tag which has a value.
  #----------------------------------------------------------------------------
  def WriteTagWithValue(self, a_Name, a_Value, **a_Attributes):
    self._WriteTag(a_Name, 0, a_Attributes)
    print >> self.i_File, ">%s</%s>" % (saxutils.escape(str(a_Value)), a_Name)

  #----------------------------------------------------------------------------
  # StartTag()
  #   Start the tag with any number of attributes; optionally end the tag if
  # desired.
  #----------------------------------------------------------------------------
  def StartTag(self, a_Name, a_MultiLineAttributes = 0, **a_Attributes):
    self._WriteTag(a_Name, a_MultiLineAttributes, a_Attributes)
    print >> self.i_File, ">"
    self.i_Tags.append(a_Name)

  #----------------------------------------------------------------------------
  # EndTag()
  #   End a tag.
  #----------------------------------------------------------------------------
  def EndTag(self):
    v_Name = self.i_Tags.pop()
    self._WriteIndentString()
    print >> self.i_File, "</%s>" % v_Name

